import logging
from typing import List, Dict

from outcache import CacheAsync

from .httpClient import HttpClient
from .models import Player, Score


class ScoreSaber:
    TIMEOUT = 10
    _url = "https://new.scoresaber.com/api"

    def __init__(self):
        self.log = logging.getLogger(__name__)

        self._http = HttpClient()

    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def start(self):
        await self._http.start()

    async def close(self):
        await self._http.close()

    async def _process_url(self, method: str, url: str) -> Dict:
        await self._http.start()

        return await self._http.request(method, url, timeout=self.TIMEOUT)

    @CacheAsync(minutes=2)
    async def _get_player_basic(self, player_id: str) -> Dict:
        return await self._process_url('GET', f"{self._url}/player/{player_id}/basic")

    async def get_player_basic(self, player_id: str) -> Player:
        response = await self._get_player_basic(player_id)

        return Player.from_dict(response["playerInfo"])

    @CacheAsync(minutes=2)
    async def _get_player_full(self, player_id: str) -> Dict:
        return await self._process_url('GET', f"{self._url}/player/{player_id}/full")

    async def get_player_full(self, player_id: str) -> Player:
        response = await self._get_player_full(player_id)

        return Player.from_dict(response["playerInfo"])

    @CacheAsync(minutes=2)
    async def _get_recent_scores(self, player_id: str, page: int = 1) -> Dict:
        return await self._process_url('GET', f"{self._url}/player/{player_id}/scores/recent/{page}")

    async def get_recent_scores(self, player_id: str, page: int = 1) -> List[Score]:
        response = await self._get_recent_scores(player_id, page)

        recent_score_list = []

        for recent_score in response["scores"]:
            recent_score_list.append(Score.from_dict(recent_score))

        return recent_score_list

    @CacheAsync(minutes=2)
    async def _get_top_scores(self, player_id: str, page: int = 1) -> Dict:
        return await self._process_url('GET', f"{self._url}/player/{player_id}/scores/top/{page}")

    async def get_top_scores(self, player_id: str, page: int = 1) -> List[Score]:
        response = await self._get_top_scores(player_id, page)

        top_score_list = []

        for top_score in response["scores"]:
            top_score_list.append(Score.from_dict(top_score))

        return top_score_list



import asyncio
import logging
from json import JSONDecodeError

import aiohttp

from .errors import RateLimitedException, NotFoundException, ServerErrorException


class HttpClient:
    WAIT_RATE_LIMIT = 60
    WAIT_SERVER_ERROR = 5

    def __init__(self):
        self.__session = None

    async def start(self):
        if self.__session is None:
            self.__session = aiohttp.ClientSession()

    async def close(self):
        if self.__session:
            await self.__session.close()

    async def request(self, method, url, **kwargs):
        async def attempt():
            async with self.__session.request(method, url, **kwargs) as response:
                self.verify_response(response)
                return await self.get_json(response)

        return await self.request_attempt(attempt)

    @staticmethod
    async def get_json(response):
        try:
            data = await response.json()
        except JSONDecodeError:
            logging.exception("JSONDecodeError, response: %r, response.text: %r", response, response.text)
            data = {"error": "Failed to decode json from scoresaber. Somethings broken."}

        return data

    @staticmethod
    def verify_response(response):
        if response.status == 429:
            raise RateLimitedException(response)

        if 400 <= response.status < 500:
            raise NotFoundException(response)

        if 500 <= response.status < 600:
            raise ServerErrorException(response)

    async def request_attempt(self, func):
        attempt = 0
        attempting = True
        last_exception = None

        while attempting and attempt < 6:
            try:
                result = await func()

                attempting = False
                return result
            except ServerErrorException as error:
                attempt += 1
                logging.info(str(error))
                logging.info(f"Waiting {self.WAIT_SERVER_ERROR} seconds...")
                last_exception = error
                await asyncio.sleep(self.WAIT_SERVER_ERROR)
            except RateLimitedException as error:
                attempt += 1
                logging.info(str(error))
                logging.info(f"Waiting {self.WAIT_RATE_LIMIT} seconds...")
                last_exception = error
                await asyncio.sleep(self.WAIT_RATE_LIMIT)

        if last_exception is not None:
            raise last_exception
